;; $Id: cell-mod.scm,v 1.2 2006/01/05 22:24:09 leavens Exp $

;;; Adapted from Figure 4.5.1 : page 125, of EOPL first edition
;;; But the typing is unsound (can't allow polymorphic storage like this).

(module cell-mod (lib "typedscm.ss" "typedscm")

  (provide cell-of cell? cell cell-ref cell-set! cell-swap!)

  ;; type predicates
  (deftype cell-of
    (forall (T) (-> ((type-predicate-for T))
                    (type-predicate-for (cell-of T)))))
  (deftype cell? (type-predicate-for (cell-of datum)))

  ;; constructor
  (deftype cell (forall (T) (-> (T) (cell-of T))))

  ;; observer
  (deftype cell-ref (forall (T) (-> ((cell-of T)) T)))

  ;; mutators
  (deftype cell-set! (forall (T) (-> ((cell-of T) T) void)))
  (deftype cell-swap! (forall (T) (-> ((cell-of T) (cell-of T)) void)))

  (defrep (forall (T) (cell-of T) (vector-of datum)))

  (define cell-tag (has-type datum "cell"))

  (define cell
    (lambda (x)
      ;; ENSURES: result is a new cell containing x
      (vector cell-tag x)))

  (define cell-of
    (forall (T?)
      (has-type-trusted
        (type-predicate-for (vector-of datum))
        (lambda (x)
          (and (cell? x)
               (T? (vector-ref x 1)))))))

  (define cell?
    (lambda (x)
      (and (vector? x)
           (= (vector-length x) 2)
           (equal? (vector-ref x 0) cell-tag))))

  (define cell-ref
    (lambda (x)
      ;; ENSURES: result is the contents of x
      (if (cell? x)
          (vector-ref x 1)
          (error "Invalid argument to cell-ref: " x))))

  (define cell-set!
    (lambda (x value)
      (if (cell? x) 
          (vector-set! x 1 value)
          (error "Illegal argument to cell-set!: " x))))

  (define cell-swap!
    (lambda (cell-1 cell-2)
      (let ((temp (cell-ref cell-1)))
        (cell-set! cell-1 (cell-ref cell-2))
        (cell-set! cell-2 temp))))

  ) ;; end module
