;; $Id: ch5-4-2.scm,v 1.3 2006/01/30 03:18:00 leavens Exp $
;;; OOPL interpreter with flat fields.

(module ch5-4-2 (lib "typedscm.ss" "typedscm")
  
(provide read-eval-print eval-program run
         scan&parse just-scan
         run-all run-functional run-oop run-one equal-external-reps?)
  
(require (lib "environment-5-4-2.scm" "lib342")
         (lib "reference-3-7.scm" "lib342")
         (lib "relax-type-checking.scm" "typedscm")
         (lib "test-suite.scm" "lib342"))

;;;;;;;;;;;;;;;; top level and tests ;;;;;;;;;;;;;;;;

(deftype run (-> (string) datum))
(define run
  (lambda (string)
    (eval-program (scan&parse string))))

(define functional-groups '(lang3-5 lang3-6 lang3-7))

(define oop-groups '(oop))

(deftype run-all (-> () void))
(define run-all
  (lambda ()
    (run-experiment run use-execution-outcome
      (append functional-groups oop-groups) (all-tests))))

(define run-functional
  (lambda ()
    (run-experiment run use-execution-outcome
      functional-groups (all-tests))))

(define run-oop
  (lambda ()
    (run-experiment run use-execution-outcome
      oop-groups (all-tests))))

(deftype run-one (-> (symbol) datum))
(define run-one
  (lambda (test-name)
    (run-test run test-name)))

;; needed for testing
(deftype equal-external-reps? (-> (datum datum) boolean))
(define equal-external-reps?
  (lambda (x y)
    (equal? x y)))

;;;;;;;;;;;;;;;; grammatical specification ;;;;;;;;;;;;;;;;

(define the-lexical-spec
  '((whitespace (whitespace) skip)
    (comment ("%" (arbno (not #\newline))) skip)
    (identifier
      (letter (arbno (or letter digit "_" "-" "?")))
      symbol)
    (number (digit (arbno digit)) number)))

(define the-grammar
  '((program ((arbno class-decl) expression) a-program)

    ;; new for oop
    (class-decl                         
      ("class" identifier 
        "extends" identifier                   
         (arbno "field" identifier)
         (arbno method-decl)
         )
      a-class-decl)

    ;; new for oop
    (method-decl
      ("method" identifier 
        "("  (separated-list identifier ",") ")" ; method formals
        expression 
        )
      a-method-decl)

    (expression (number) lit-exp)
    (expression ("quote" identifier) quote-exp)
    (expression (identifier) var-exp)
    (expression
      (primitive "(" (separated-list expression ",") ")")
      primapp-exp)
    (expression
      ("if" expression "then" expression "else" expression)
      if-exp)
    (expression
      ("let" (arbno  identifier "=" expression) "in" expression)
      let-exp)
    (expression
      ("proc" "(" (separated-list identifier ",") ")" expression)
      proc-exp)
    (expression
      ("(" expression (arbno expression) ")")
      app-exp)
    (expression
      ("begin" expression (arbno ";" expression) "end")
      begin-exp)
    (expression                         
      ("letrec"
        (arbno identifier "(" (separated-list identifier ",") ")"
          "=" expression)
        "in" expression)
      letrec-exp)
    (expression ("set" identifier "=" expression) varassign-exp)
    ;; new for oop
    (expression 
      ("new" identifier "(" (separated-list expression ",") ")")
      new-object-exp)
    (expression
      ("send" expression identifier
        "("  (separated-list expression ",") ")")
      method-app-exp)
    (expression                                
      ("super" identifier    "("  (separated-list expression ",") ")")
      super-call-exp)

    (primitive ("+")     add-prim)
    (primitive ("-")     subtract-prim)
    (primitive ("*")     mult-prim)
    (primitive ("add1")  incr-prim)
    (primitive ("sub1")  decr-prim)
    (primitive ("zero?") zero-test-prim)
    (primitive ("equal?") equal-prim)
    (primitive ("greater?") greater-prim)
    (primitive ("less?") less-prim)
    (primitive ("list") list-prim)
    (primitive ("cons") cons-prim)
    (primitive ("nil")  nil-prim)
    (primitive ("car")  car-prim)
    (primitive ("cdr")  cdr-prim)
    (primitive ("null?") null?-prim)
    ))

(deftype show-the-datatypes (-> () (list-of datum)))
(define show-the-datatypes
  (lambda ()
    (sllgen:make-define-datatypes the-lexical-spec the-grammar)
    (sllgen:list-define-datatypes the-lexical-spec the-grammar)))

;; The following is generated by calling show-the-datatypes
;; (and then removing the outer lists, indenting, and renaming fields).
(define-datatype
  program
  program?
  (a-program
   (class-decls (list-of class-decl?))
   (exp expression?)))

(define-datatype
  class-decl
  class-decl?
  (a-class-decl
   (class-name symbol?)
   (super-name symbol?)
   (field-ids (list-of symbol?))
   (method-decls (list-of method-decl?))))

(define-datatype
  method-decl
  method-decl?
  (a-method-decl
   (method-name symbol?)
   (ids (list-of symbol?))
   (body expression?)))

(define-datatype
  expression
  expression?
  (lit-exp (datum number?))
  (quote-exp (sym symbol?))
  (var-exp (id symbol?))
  (primapp-exp
   (prim primitive?)
   (rands (list-of expression?)))
  (if-exp
   (test-exp expression?)
   (true-exp expression?)
   (false-exp expression?))
  (let-exp
   (ids (list-of symbol?))
   (rands (list-of expression?))
   (body expression?))
  (proc-exp
   (ids (list-of symbol?))
   (body expression?))
  (app-exp
   (rator expression?)
   (rands (list-of expression?)))
  (begin-exp
   (first expression?)
   (rest (list-of expression?)))
  (letrec-exp
   (proc-names (list-of symbol?))
   (idss (list-of (list-of symbol?)))
   (bodies (list-of expression?))
   (letrec-body expression?))
  (varassign-exp
   (id symbol?)
   (rhs-exp expression?))
  (new-object-exp
   (class-name symbol?)
   (rands (list-of expression?)))
  (method-app-exp
   (obj-exp expression?)
   (method-name symbol?)
   (rands (list-of expression?)))
  (super-call-exp
   (method-name symbol?)
   (rands (list-of expression?)))
  )

(define-datatype
  primitive
  primitive?
  (add-prim)
  (subtract-prim)
  (mult-prim)
  (incr-prim)
  (decr-prim)
  (zero-test-prim)
  (equal-prim)
  (greater-prim)
  (less-prim)
  (list-prim)
  (cons-prim)
  (nil-prim)
  (car-prim)
  (cdr-prim)
  (null?-prim)
  )

(deftype scan&parse (-> (string) program))
(define scan&parse
  (lambda (ptext)
    ((sllgen:make-string-parser the-lexical-spec the-grammar)
     ptext)))

(deftype just-scan (forall (TOK) (-> (string) (list-of TOK))))
(define just-scan
  (lambda (ptext)
    ((sllgen:make-string-scanner the-lexical-spec the-grammar)
     ptext)))

(deftype read-eval-print (-> () poof))
(define read-eval-print
  (lambda ()
    ((sllgen:make-rep-loop
      "--> "
      (lambda (pgm) (eval-program pgm))
      (sllgen:make-stream-parser 
       the-lexical-spec
       the-grammar)))))

;;;;;;;;;;;;;;;; AST (declaration) helpers ;;;;;;;;;;;;;;;;

(deftype class-decl->class-name (-> (class-decl) symbol))
(define class-decl->class-name
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        class-name))))

(deftype class-decl->super-name  (-> (class-decl) symbol))
(define class-decl->super-name
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        super-name))))

(deftype class-decl->field-ids  (-> (class-decl) (list-of symbol)))
(define class-decl->field-ids
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        field-ids))))

(deftype class-decl->method-decls (-> (class-decl) (list-of method-decl)))
(define class-decl->method-decls
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        m-decls))))

(deftype method-decl->method-name (-> (method-decl) symbol))
(define method-decl->method-name
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) method-name))))

(deftype method-decl->ids (-> (method-decl) (list-of symbol)))
(define method-decl->ids
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) ids))))

(deftype method-decl->body (-> (method-decl) expression))
(define method-decl->body
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) body))))

(deftype method-decls->method-names
  (-> ((list-of method-decl)) (list-of symbol)))
(define method-decls->method-names
  (lambda (mds)
    (map method-decl->method-name mds)))

;;; Domains for this interpreter:
;;;
;;; Expressed-Value = Number + ProcVal + List(Expressed-Value) + Obj + Symbol
;;; Denoted-Value = Ref(Expressed-Value)

;;; ADTs used

;;;;;;;;;;;;;;;; ProcVal ;;;;;;;;;;;;;;;;

(deftype procval? (type-predicate-for procval))
(deftype closure (-> ((list-of symbol) expression environment) procval))
(deftype apply-procval
  (-> (procval (list-of Expressed-Value)) Expressed-Value))

(define-datatype procval procval?
  (closure 
    (ids (list-of symbol?)) 
    (body expression?)
    (env environment?)))

(define apply-procval
  (lambda (proc args)
    (cases procval proc
      (closure (ids body env)
        (eval-expression body (extend-env ids args env))))))

;;;;;;;;;;;;;;;; classes ;;;;;;;;;;;;;;;;

;;; classes are represented by their class-decls.

(define class? class-decl?)             ; not used

;;;;;;;;;;;;;;;; objects ;;;;;;;;;;;;;;;;

;; An object is now just a single part, with a vector representing the
;; managed storage for the all the fields. 

(define-datatype object object? 
  (an-object
    (class-name symbol?)
    (fields vector?)))

(deftype new-object (-> (symbol) object))
(define new-object
  (lambda (class-name)
    (an-object
      class-name
      (make-vector
       (roll-up-field-length class-name)
       (number->expressed 0)))))

(deftype roll-up-field-length (-> (symbol) number))
(define roll-up-field-length
  (lambda (class-name)
    (if (eqv? class-name 'object)
      0
      (+
        (roll-up-field-length
          (class-name->super-name class-name))
        (length (class-name->field-ids class-name))))))

;;;;;;;;;;;;;;;; methods ;;;;;;;;;;;;;;;;

;;; Methods are represented by their declarations.  They are closed
;;; over their fields at application time, by apply-method.

(deftype find-method-and-apply
  (-> (symbol symbol object (list-of Expressed-Value)) Expressed-Value))
(define find-method-and-apply
  (lambda (m-name host-name self args)
    (if (eqv? host-name 'object)
      (eopl:error 'find-method-and-apply
        "No method for name ~s" m-name)
      (let ((declared-methods (class-name->method-decls host-name)))
        (if (has-method-decl? m-name declared-methods)
            (let ((m-decl (lookup-method-decl m-name declared-methods)))
              (apply-method m-decl host-name self args))
            (find-method-and-apply
             m-name (class-name->super-name host-name)
             self args))))))

(deftype apply-method
  (-> (method-decl symbol object (list-of Expressed-Value)) Expressed-Value))
(define apply-method
  (lambda (m-decl host-name self args)
    (let ((ids (method-decl->ids m-decl))
          (body (method-decl->body m-decl))
          (super-name (class-name->super-name host-name))
          (field-ids (roll-up-field-ids host-name))
          (fields (object->fields self)))
      (eval-expression body
        (extend-env
          (cons '%super (cons 'self ids))
          (cons (symbol->expressed super-name)
                (cons (object->expressed self) args))
          (extend-env-refs field-ids fields (empty-env)))))))

(deftype roll-up-field-ids (-> (symbol) (list-of symbol)))
(define roll-up-field-ids
  (lambda (class-name)
    (if (eqv? class-name 'object)
      '()
      (append
        (roll-up-field-ids
          (class-name->super-name class-name))
        (class-name->field-ids class-name)))))

;;;;;;;;;;;;;;;; method environments ;;;;;;;;;;;;;;;;

(deftype has-method-decl? (-> (symbol (list-of method-decl)) boolean))
(define has-method-decl?
  (lambda (m-name m-decls)
    (and (not (null? m-decls))
         (or (eqv? m-name (method-decl->method-name (car m-decls)))
             (has-method-decl? m-name (cdr m-decls))))))

(deftype lookup-method-decl (-> (symbol (list-of method-decl)) method-decl))
(define lookup-method-decl 
  (lambda (m-name m-decls)
    (cond
      ((null? m-decls)
       (eopl:error 'lookup-method-decl "No method named ~s" m-name))
      ((eqv? m-name (method-decl->method-name (car m-decls)))
       (car m-decls))
      (else (lookup-method-decl m-name (cdr m-decls))))))
      
;;;;;;;;;;;;;;;; class environments ;;;;;;;;;;;;;;;;

;;;; we'll just use the list of class-decls.

(deftype the-class-env (list-of class-decl))
(define the-class-env '())

(deftype elaborate-class-decls! (-> ((list-of class-decl)) void))
(define elaborate-class-decls!
  (lambda (c-decls)
    (set! the-class-env c-decls)))

(deftype lookup-class (-> (symbol) class-decl))
(define lookup-class
  (lambda (name)
    (let loop ((env the-class-env))
      (cond
        ((null? env) (eopl:error 'lookup-class "Unknown class ~s" name))
        ((eqv? (class-decl->class-name (car env)) name) (car env))
        (else (loop (cdr env)))))))

;;;;;;;;;;;;;;;; selectors of all sorts ;;;;;;;;;;;;;;;;

(deftype object->fields (-> (object) (vector-of Expressed-Value)))
(define object->fields
  (lambda (obj)
    (cases object obj
      (an-object (class-decl fields)
        fields))))

(deftype object->fields-ids (-> (object) (list-of symbol)))
(define object->field-ids
  (lambda (object)
    (class-decl->field-ids
      (object->class-decl object))))

(deftype object->class-decl (-> (object) class-decl))
(define object->class-decl
  (lambda (obj)
    (lookup-class (object->class-name obj))))

(deftype object->class-name (-> (object) symbol))
(define object->class-name
  (lambda (obj)
    (cases object obj
      (an-object (class-name fields)
        class-name))))

(deftype class-name->field-ids (-> (symbol) (list-of symbol)))
(define class-name->field-ids
  (lambda (class-name)
    (class-decl->field-ids (lookup-class class-name))))

(deftype class-name->method-decls (-> (symbol) (list-of method-decl)))
(define class-name->method-decls
  (lambda (class-name)
    (class-decl->method-decls (lookup-class class-name))))

(deftype class-name->super-name (-> (symbol) symbol))
(define class-name->super-name
  (lambda (class-name)
    (class-decl->super-name (lookup-class class-name))))

;;; Anonymous figure : page 80

(deftype true-value? (-> (Expressed-Value) boolean))
(define true-value?
  (lambda (x)
    (not (zero? (expressed->number x)))))

(deftype boolean->expressed (-> (boolean) Expressed-Value))
(define boolean->expressed
  (lambda (b)
    (number->expressed
     (if b 1 0))))

;;;;;;;;; Expressed-Value ;;;;;;;;;;;;;;;;;;;;;;;;;

;; upcasts
(deftype number->expressed (-> (number) Expressed-Value))
(deftype procval->expressed (-> (Procval) Expressed-Value))
(deftype list->expressed (-> ((list-of Expressed-Value)) Expressed-Value))
(deftype object->expressed (-> (object) Expressed-Value))
(deftype symbol->expressed (-> (symbol) Expressed-Value))

;; downcasts
(deftype expressed->number (-> (Expressed-Value) number))
(deftype expressed->procval (-> (Expressed-Value) Procval))
(deftype expressed->list (-> (Expressed-Value) (list-of Expressed-Value)))
(deftype expressed->object (-> (Expressed-Value) object))
(deftype expressed->symbol (-> (Expressed-Value) symbol))

;; debugging
(deftype expressed->printable (-> (Expressed-Value) datum))

;; tests
(deftype number->expressed? (-> (Expressed-Value) boolean))
(deftype procval->expressed? (-> (Expressed-Value) boolean))
(deftype list->expressed? (-> (Expressed-Value) boolean))
(deftype object->expressed? (-> (Expressed-Value) boolean))
(deftype symbol->expressed? (-> (Expressed-Value) boolean))

(define-datatype Expressed-Value expval?
  (number->expressed (num number?))
  (procval->expressed (pv procval?))
  (list->expressed (lst (list-of expval?)))
  (object->expressed (obj object?))
  (symbol->expressed (sym symbol?)))

(define expressed->number
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) num)
      (else (error "expressed->number passed non-number argument: " ev)))))

(define expressed->procval
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) pv)
      (else (error "expressed->procval passed non-procval argument: " ev)))))

(define expressed->list
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) lst)
      (else (error "expressed->list passed non-list argument: " ev)))))

(define expressed->object
  (lambda (ev)
    (cases Expressed-Value ev
      (object->expressed (obj) obj)
      (else (error "expressed->object passed non-object argument: " ev)))))

(define expressed->symbol
  (lambda (ev)
    (cases Expressed-Value ev
      (symbol->expressed (sym) sym)
      (else (error "expressed->symbol passed non-symbol argument: " ev)))))

(define expressed->printable
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) (has-type datum num))
      (procval->expressed (pv)
        (cases procval pv
          (closure (ids body env)
           (has-type datum (list 'closure ids body "<env>")))))
      (list->expressed (lst)
         (has-type datum (map expressed->printable lst)))
      (object->expressed (obj) (has-type datum obj))
      (symbol->expressed (sym) (has-type datum sym))
      )))

(define number->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) #t)
      (else #f))))

(define procval->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) #t)
      (else #f))))

(define list->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) #t)
      (else #f))))

(define object->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (object->expressed (obj) #t)
      (else #f))))

(define symbol->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (symbol->expressed (sym) #t)
      (else #f))))


;; adaptation of environment-5-4-2 to this file
(deftype extend-env-recursively
  (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
       environment)
      environment))
(define extend-env-recursively
  (extend-env-recursively-maker
   (lambda (ids body env)
     (procval->expressed (closure ids body env)))
   (procval->expressed (closure '() (lit-exp 0) (empty-env)))))

;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

(deftype eval-program (-> (program) datum))
(define eval-program 
  (lambda (pgm)
    (cases program pgm
      (a-program (c-decls body)
        (elaborate-class-decls! c-decls)
        (expressed->printable (eval-expression body (init-env)))))))

(deftype eval-expression (-> (expression environment) Expressed-Value))
(define eval-expression 
  (lambda (exp env)
    (cases expression exp
      (lit-exp (datum)
        (number->expressed datum))
      (quote-exp (sym)
        (symbol->expressed sym))
      (var-exp (id)
        (apply-env env id))
      (primapp-exp (prim rands)
        (let ((args (eval-rands rands env)))
          (apply-primitive prim args)))
      (if-exp (test-exp true-exp false-exp)
        (if (true-value? (eval-expression test-exp env))
          (eval-expression true-exp env)
          (eval-expression false-exp env)))
      (begin-exp (exp1 exps)
        (let loop ((acc (eval-expression exp1 env))
                   (exps exps))
          (if (null? exps) acc
            (loop (eval-expression (car exps) env) (cdr exps)))))
      (let-exp (ids rands body)
        (let ((args (eval-rands rands env)))
          (eval-expression
           body
           (extend-env ids args env))))
      (proc-exp (ids body) (procval->expressed (closure ids body env)))
      (app-exp (rator rands)
        (let ((proc (eval-expression rator env))
              (args (eval-rands rands env)))
          (if (procval->expressed? proc)
            (apply-procval (expressed->procval proc) args)
            (eopl:error 'eval-expression
              "Attempt to apply non-procedure ~s" proc))))
      (letrec-exp (proc-names idss bodies letrec-body)
        (eval-expression letrec-body
          (extend-env-recursively proc-names idss bodies env)))     
      (varassign-exp (id rhs-exp)
        (begin
          (setref! 
            (apply-env-ref env id)
            (eval-expression rhs-exp env))
          (number->expressed 1)))
      (new-object-exp (class-name rands)
        (let ((args (eval-rands rands env))
              (obj (new-object class-name)))
          (ignore
           (find-method-and-apply
            'initialize class-name obj args))
          (object->expressed obj)))
      (method-app-exp (obj-exp method-name rands)
        (let ((args (eval-rands rands env))
              (obj (expressed->object (eval-expression obj-exp env))))
          (find-method-and-apply
            method-name (object->class-name obj) obj args)))
      (super-call-exp (method-name rands)
        (let ((args (eval-rands rands env))
              (obj (expressed->object (apply-env env 'self))))
          (find-method-and-apply
            method-name
            (expressed->symbol (apply-env env '%super))
            obj
            args)))
      (else (eopl:error 'eval-expression "Missing case:~s" exp))
      )))

(deftype eval-rands
  (-> ((list-of expression) environment) (list-of Expressed-Value)))
(define eval-rands
  (lambda (rands env)
    (map (lambda (x) (eval-expression x env)) rands)))

(deftype apply-primitive
  (-> (primitive (list-of Expressed-Value)) Expressed-Value))
(define apply-primitive
  (lambda (prim args)
    (define (check arity)
      (if (not (= arity (length args)))
          (eopl:error 'apply-primitive
                 "wrong number of arguments to primitive ~s"
                 prim)))
    (cases primitive prim
      (add-prim ()
        (check 2)
        (number->expressed (+ (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (subtract-prim ()
        (check 2)
        (number->expressed (- (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (mult-prim ()
        (check 2)
        (number->expressed (* (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (incr-prim ()
        (check 1)
        (number->expressed (+ (expressed->number (car args)) 1)))
      (decr-prim ()
        (check 1)
        (number->expressed (- (expressed->number (car args)) 1)))
      (zero-test-prim()
        (check 1)
        (boolean->expressed (zero? (expressed->number (car args)))))
      (equal-prim ()
        (check 2)
        (boolean->expressed
         (cases Expressed-Value (car args)
           (number->expressed (n1)
             (= n1 (expressed->number (cadr args))))
           (symbol->expressed (s1)
             (eq? s1 (expressed->symbol (cadr args))))
           (else (eopl:error
                  'apply-primitive
                  "Can only compare numbers and symbols with equal?, not ~s"
                  (car args))))))
      (greater-prim () 
        (check 2)
        (boolean->expressed (> (expressed->number (car args))
                               (expressed->number (cadr args)))))
      (less-prim () 
        (check 2)
        (boolean->expressed (< (expressed->number (car args))
                               (expressed->number (cadr args)))))
      (list-prim ()
        ;; any number of arguments is ok
        (list->expressed args))
      (nil-prim ()
        (check 0)
        (list->expressed '()))
      (car-prim () 
        (check 1)
        (car (expressed->list (car args))))
      (cdr-prim ()
        (check 1)
        (list->expressed (cdr (expressed->list (car args)))))
      (cons-prim () 
        (check 2)
        (list->expressed (cons (car args) (expressed->list (cadr args)))))
      (null?-prim ()
        (check 1)
        (boolean->expressed (null? (expressed->list (car args)))))
      )))

(deftype init-env (-> () environment))
(define init-env 
  (lambda ()
    (empty-env)))

) ;; end module
