;;; $Id: combinator-lambda-1-exp.scm,v 1.4 2006/02/14 03:56:42 leavens Exp $

;;; An ADT for the untyped lambda 1-argument calculus with
;;; the three constants LS, LK, and LI, representing the combinators
;;; S, K, and I.

;;; Grammar:
;;;  <expression>
;;;              ::= <identifier>
;;;                | (lambda (<identifier>) <expression>)
;;;                | (<expression> <expression>)
;;;                | LS | LK | LI

(module combinator-lambda-1-exp (lib "typedscm.ss" "typedscm")
  
(provide expression expression? var-exp lambda-exp app-exp LS LK LI
         parse-expression unparse-expression)
         
(define-datatype expression expression?
  (var-exp (id symbol?))
  (lambda-exp (formal symbol?) (body expression?))
  (app-exp (rator expression?) (rand expression?))
  (LS)
  (LK)
  (LI))

(deftype parse-expression (-> (datum) expression))
(define parse-expression
  (lambda (exp)
    (test-type
      expression?
      (cond
       ((symbol? exp)
        (case exp
          ((S s) (LS)) ;; DrScheme can be case sensitive...
          ((K k) (LK))
          ((I i) (LI))
          (else (var-exp exp))))
       ((and (list? exp) (= (length exp) 3) (eq? (car exp) 'lambda))
        (lambda-exp (caadr exp)
                    (parse-expression (caddr exp))))
       ((and (list? exp) (= (length exp) 2))
        (app-exp (parse-expression (car exp))
                 (parse-expression (cadr exp))))
       (else (error "parse-expression: bad syntax in expression:"
                    exp))))))

(deftype unparse-expression (-> (expression) datum))
(define unparse-expression
  (lambda (exp)
    (cases expression exp
      (LS () (has-type datum 'S))
      (LK () (has-type datum 'K))
      (LI () (has-type datum 'I))
      (var-exp (id) (has-type datum id))
      (lambda-exp (formal body)
        (list 'lambda (list formal) (unparse-expression body)))
      (app-exp (rator rand) (list (unparse-expression rator)
                                  (unparse-expression rand)))
      )))

) ;; end module
