;;; $Id: phone-book-as-ribcage.scm,v 1.4 2006/01/05 22:24:09 leavens Exp $
;;;---------------------------------------------------------------------
;;; An ADT for a simple phone book
;;;---------------------------------------------------------------------

;;;---------------------------------------------------------------------
;;; The declarations for the public (client-visible) interface of the
;;; phone book ADT
;;;---------------------------------------------------------------------

(module phone-book-as-ribcage (lib "typedscm.ss" "typedscm")

(provide create-empty-phone-book add-to-phone-book look-up
         add-several-to-phone-book)

(deftype create-empty-phone-book (-> () phone-book))
(deftype add-to-phone-book (-> (symbol number phone-book) phone-book))
(deftype look-up (-> (phone-book symbol) number))


;;;---------------------------------------------------------------------
;;; A ribcage representation of the phone book ADT
;;;---------------------------------------------------------------------

(require (lib "ribassoc-mod.scm" "lib342"))

(defrep (phone-book (list-of (pair-of (list-of symbol)
                                     (vector-of number)))))
(define create-empty-phone-book
  (lambda () '()))

(define add-to-phone-book
  (lambda (name number old-book)
    (cons (make-rib (list name) (list number))
          old-book)))

(deftype make-rib (forall (T) 
                          (-> ((list-of symbol) (list-of T))
                              (pair-of (list-of symbol) (vector-of T)))))
(define make-rib
  (lambda (ls lt)
    (cons ls (list->vector lt))))

(define look-up
  (lambda (book who)
    (if (null? book)
        (error "no listing for" who)
        (let ((first-listings (car book))
              (old-book (cdr book)))
          (let ((names (car first-listings))
                (numbers (cdr first-listings)))
            (let ((result
                   (test-type datum? (ribassoc who names numbers '*fail*))))
              (if (eq? result '*fail*)
                  (look-up old-book who)
                  (test-type number? result))))))))

;;;---------------------------------------------------------------------
;;; With the ribcage rep it is easy to add a procedure for multiple
;;; names and numbers to the ADT
;;;---------------------------------------------------------------------

(deftype add-several-to-phone-book
  (-> ((list-of symbol) (list-of number) phone-book) phone-book))
(define add-several-to-phone-book
  (lambda (names numbers old-book)
    (cons (make-rib names numbers)
          old-book)))

) ;; end module       
