;; $Header: /home/larch/research/master/typedscm/Attic/ch5-7-1-let.scm,v 1.1 1997/07/25 15:07:49 ruby Exp $

;; Code for EOPL section 5.7.1 (dynamic scoping)
;; using the chararacter string parser.

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;	            | proc <varlist> <exp>
;;	            | <var> := <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*

;;; Syntactic sugars
;;       see the file below

;(load-from-lib "syntax-expand-let.scm")
(load-from-lib "syntax-expand-0.scm")

;;; Top-level Forms
;;      <form> ::= <exp>

(load-from-lib "eval-form-0.scm")

;;; Parsing

;; To use "run", you also need to do
;;        (load "run.scm")
;; To use "read-eval-print", you also need to do
;;        (load "rep.scm")
;; Note that you can't use both of these at the same time.

(load-from-lib "undefine-read-eval-print.scm")
(load-from-lib "undefine-run.scm")
(load-from-lib "ch5-char-parser.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Cell(Expressed-Value)
;;; Expressed-Value = Number + Procedure-Text
;;;                   + Pair(Expressed-Value Expressed-Value) + Null + Void
;;; Procedure-Text = prim-proc + proc

;;; ADTs used

(load-from-lib "environments.scm")
(load-from-lib "cell.scm")

;;; Helpful stuff

(load-from-lib "compose.scm")

;;; Truth values

;; Anonymous figure : page 146

(define true-value?
  ; TYPE: (-> (integer) boolean)
  (lambda (x)
    (not (zero? x))))

(define boolean->int
  ; TYPE: (-> (boolean) integer)
  (lambda (x)
    (if  x  1  0)))

;;; Primitive procedures

;; Anonymous figure : page 142, 143

(define-record prim-proc (prim-op))

(define apply-prim-op
  ; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (+ (car args) (cadr args)))
      ((-) (- (car args) (cadr args)))
      ((*) (* (car args) (cadr args)))
      ((add1) (+ (car args) 1))
      ((sub1) (- (car args) 1))
      ; primitive operators added to interpreter follow
      ((minus) (- (car args)))
      ((cons) (cons (car args) (cadr args)))
      ((car) (car (car args)))
      ((cdr) (cdr (car args)))
      ((cadr) (cadr (car args)))
      ((caddr) (caddr (car args)))
      ((cadddr) (cadddr (car args)))      
      ((list) args)  
      ((null) (boolean->int (null? (car args))) )
      ((eq) (boolean->int (eq? (car args) (cadr args))) ) 
      ((equal) (boolean->int (= (car args) (cadr args))) )
      ((less) (boolean->int (< (car args) (cadr args))) ) 
      ((greater) (boolean->int (> (car args) (cadr args))) )
      ((zero) (boolean->int (zero? (car args))) )
      ((write) (write (car args)))
      ((print) (apply displayln args))
      ((writeln) (begin (for-each (lambda (arg)
				    (write arg)
				    (display " "))
				  args)
			(newline)))
      ((space) (display " "))
      ((newline) (newline))
      ((printchar) (display (integer->char (car args))))
      (else (error "Invalid prim-op name:" prim-op)))))

(define prim-op-names  ; TYPE: (list symbol)
  ; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar))

(define init-env       ; TYPE: Environment
  (extend-env
    prim-op-names
    (map (compose make-cell make-prim-proc) prim-op-names)
    ; modified to add the empty list
    (extend-env
      '(emptylist)
      (list (make-cell '()))
      the-empty-env)))

;;; Evaluation of expressions

;;; Figure 5.7.1 : page 169

(define eval-exp
  ; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (exp env)
    (variant-case exp
      (lit (datum) datum)
      (varref (var) (cell-ref (apply-env env var)))
      (app (rator rands)
           (let ((proc (eval-exp rator env))
		 (args (eval-rands rands env)))
	     (apply-proc proc args env)))
      (if (test-exp then-exp else-exp)
	  (if (true-value? (eval-exp test-exp env))
	      (eval-exp then-exp env)
	      (eval-exp else-exp env)))
      (proc (formals body) exp)
      (let (decls body) 
	(let ((vars (map decl->var decls))
	      (exps (map decl->exp decls)))
          (let ((new-env (extend-env
			   vars
			   (map (lambda (e) 
				  (make-cell (eval-exp e env))) exps)
                           env)))
            (eval-exp body new-env))))
      (varassign (var exp)
		 (cell-set! (apply-env env var) (eval-exp exp env)))
      (else (error "Invalid abstract syntax: " exp)))))

(define apply-proc
  ; TYPE: (-> (Procedure-Text (list Expressed-Value) Environment)
  ;           Expressed-Value)
  (lambda (proc args current-env)
    (variant-case proc
      (prim-proc (prim-op) (apply-prim-op prim-op args))
      (proc (formals body)
        (eval-exp body 
          (extend-env formals (map make-cell args) current-env)))
      (else (error "Invalid procedure:" proc)))))

(define eval-rands
  ; TYPE: (-> ((list parsed-exp) Environment) (list Expressed-Value))
  (lambda (rands env)
    (map (lambda (rand) (eval-exp rand env))
         rands)))

;;; something to make the read-eval-print loop work...

(define closure?
  ; TYPE: (-> (datum) boolean)
  (lambda (x) #f))
