;; $Header: /home/larch/research/master/typedscm/Attic/recursion-examples.scm,v 1.2 1996/02/10 10:39:02 leavens Exp $
;;; Name:  Gary T. Leavens
;;; Section:  all

;;; various examples from section on recursion

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Examples:
;; (num-of 'a 'a) ==> 1
;; (num-of 'a 'b) ==> 0
;; (num-of 'x '(x y (x))) ==> 2
;; (num-of 'x '(y (x))) ==> 1

;; Grammar:
;; <symbol-expression>::= <symbol> | <s-list>
;; <s-list> ::= ( {<symbol-expression>}* )

(define num-of
  ; TYPE: (-> (symbol symbol-expression) number)
  (lambda (sym symexp)
    (if (symbol? symexp)
	(if (eq? sym symexp)
	    1
	    0)
	(num-of-slist sym symexp))))


;; Examples:
;; (num-of-slist 'x '()) ==> 0
;; (num-of-slist 'x '(x y (x))) ==> 2
;; (num-of-slist 'x '(y (x))) ==> 1

(define num-of-slist
  ; TYPE: (-> (symbol s-list) number)
  (lambda (sym slist)
    (if (null? slist)
	0
	(+ (num-of sym (car slist))
	   (num-of-slist sym (cdr slist))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;

; Examples:
; (fact 5) = 120
; (fact 4) = 24
; (fact 3) = 6
; (fact 0) = 1

(define fact
  ; TYPE: (-> (number) (number))
  (lambda (n)
    (if (zero? n)
	1
	(* n (fact (sub1 n))))))

(define sub1
  ; TYPE: (-> (number) number)
  (lambda (n)
    (- n 1)))

;;;;;;;;;;;;;;;;;;;;;;

;; Examples
; (add1-to-each '()) ==> ()
; (add1-to-each '(4 5 6)) ==> (5 6 7)

(define add1-to-each
  ; TYPE: (-> ((list number)) (list number))
  (lambda (lon)
    (if (null? lon)
	'()
	(cons (+ 1 (car lon))
	      (add1-to-each (cdr lon))))))

;;;;;;;;;;;;;;;;;;;;;

; for this example, see also pages 48-49 of EOPL

;; Examples:
;; (subst 'x 'y '()) ==> ()
;; (subst 'k 'a '(a b c d a)) ==> (k b c d k)
;; (subst 'b 'a '((a a) (c))) ==> ((b b) (c))
;; (subst 'k 'a '(a ((b () a)) ((a ()))))
;;           ==> (k ((b () k)) ((k ())))

;; Grammar:
;; <symbol-expression>::= <symbol> | <s-list>
;; <s-list> ::= ( {<symbol-expression>}* )

(define subst
  ; TYPE: (-> (symbol symbol s-list) s-list)
  (lambda (new old slist)
    (if (null? slist)
	'()
	(cons (subst-sym-exp new old (car slist))
	      (subst new old (cdr slist))))))

;; Examples:
;; (subst-sym-exp 'k 'a 'a) ==> k
;; (subst-sym-exp 'k 'a 'b) ==> b
;; (subst-sym-exp 'x 'y '()) ==> ()
;; (subst-sym-exp 'k 'a '(a b c d a)) ==> (k b c d k)
;; (subst-sym-exp 'b 'a '((a a) (c))) ==> ((b b) (c))
;; (subst-sym-exp 'k 'a '(a ((b () a)) ((a ()))))
;;                   ==> (k ((b () k)) ((k ())))

(define subst-sym-exp
  ; TYPE: (-> (symbol symbol symbol-expression) symbol-expression)
  (lambda (new old symexp)
    (if (symbol? symexp)
	(if (eq? old symexp)
	    new
	    symexp)
	(subst new old symexp))))

;;;;;;;;;;;;;;;;;;;;;;;;;

;; Grammar:
; <exp> ::= <varref>
;	  | (lambda (<var>) <exp>)
;	  | (<exp> <exp>)
; <varref> ::= <var>
; <var> ::= <symbol>

;; Examples:
; (count-varrefs 'x) ==> 1
; (count-varrefs '(x y)) ==> 2
; (count-varrefs '(lambda (x) x)) ==> 1
; (count-varrefs '(f (lambda (x) x))) ==> 2

(define count-varrefs
  ; TYPE: (-> (lambda-exp) number)
  (lambda (exp)
    (if (varref? exp)
	1
	(if (lambda? exp)
	    (count-varrefs (lambda->body exp))
	    ; application
	    (+ (count-varrefs (app->rator exp))
	       (count-varrefs (app->rand exp)))))))

; helping procedures for the above:

;; Examples:
; (varref? 'x) ==> #t
; (varref? '(lambda (x) x)) ==> #f
; (varref? '(car ls)) ==> #f

(define varref?
  ; TYPE: (-> (lambda-exp) boolean)
  symbol?)

;; Examples:
; (lambda? 'x) ==> #f
; (lambda? '(lambda (x) x)) ==> #t
; (lambda? '(car ls)) ==> #f

(define lambda?
  ; TYPE: (-> (lambda-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (eq? (car exp) 'lambda))))

;; Examples:
; (lambda->body '(lambda (x) x)) ==> x
; (lambda->body '(lambda (x) (e x))) ==> (e x)

(define lambda->body
  ; TYPE: (-> (lambda-exp) lambda-exp)
  ; REQUIRES: the argument has the form (lambda (<var>) <exp>)
  caddr)

;; Examples:
; (app->rator '(e x) ==> e

(define app->rator
  ; TYPE: (-> (lambda-exp) lambda-exp)
  ; REQUIRES: the argument has the form (<exp> <exp>)
  car)

;; Examples:
; (app->rand '(e x) ==> x

(define app->rand
  ; TYPE: (-> (lambda-exp) lambda-exp)
  ; REQUIRES: the argument has the form (<exp> <exp>)
  cadr)
