;;; $Id: tc-char-seq-as-stream.scm,v 1.5 2006/02/06 01:09:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Character sequences, represented by streams

(module tc-char-seq-as-stream (lib "typedscm.ss" "typedscm")
  
(provide tc:char-seq? tc:the-null-char-seq tc:char-seq-null?
         tc:make-char-seq tc:char-seq-head tc:char-seq-tail 
         tc:string->char-seq tc:stream->char-seq)

(deftype tc:char-seq? (type-predicate-for tc:char-seq))
(deftype tc:the-null-char-seq (-> () tc:char-seq))
(deftype tc:char-seq-null? (-> (tc:char-seq) boolean))
(deftype tc:make-char-seq (-> (char tc:char-seq) tc:char-seq))
(deftype tc:char-seq-head (-> (tc:char-seq) char))
(deftype tc:char-seq-tail (-> (tc:char-seq) tc:char-seq))
(deftype tc:string->char-seq (-> (string) tc:char-seq))
(deftype tc:stream->char-seq (-> ((tc:stream-of char)) tc:char-seq))

(require (lib "tc-parsing-stream.scm" "typedscm"))

(defrep (tc:char-seq (tc:stream-of char)))

(define tc:char-seq?
  (lambda (d)
    ((tc:stream-of char?) d)))

(define tc:the-null-char-seq
  (lambda ()
    (tc:the-null-stream)))

(define tc:char-seq-null?
  (lambda (cs)
    (tc:stream-null? cs)))

(define tc:make-char-seq
  (lambda (ch seq)
    (tc:make-stream ch (lambda () seq))))

(define tc:char-seq-head
  (lambda (seq)
    (tc:stream-car seq)))

(define tc:char-seq-tail
  (lambda (seq)
    (tc:stream-cdr seq)))

(define tc:string->char-seq
  (lambda (str)
    (let ((len (string-length str)))
    (if (= 0 len)
	(tc:the-null-stream)
	(tc:make-stream
	 (string-ref str 0)
	 (lambda ()
	   (tc:string->char-seq (substring str 1 len))))))))

(define tc:stream->char-seq
  (lambda (stream)
    stream))
)
