;;; $Id: tc-hash-table.tst,v 1.7 2006/02/10 05:23:31 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

(require (lib "tc-hash-table.scm" "typedscm"))

(define emptyhash (tc:make-hash-table))
(define a3 (tc:hash-table-associate 'a 3 emptyhash))
(define b4c5a3 (tc:hash-table-associate* '(b c) '(4 5) a3))

(run-regression-tests-equal?
 `(
    ( (not (not (tc:hash-table-defined? 'a emptyhash)))
      ==> #f )

    ( (not (not (tc:hash-table-defined? 'a a3)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'b a3)))
      ==> #f )

    ( (not (not (tc:hash-table-defined? 'b b4c5a3)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'c b4c5a3)))
      ==> #t )

    ( (tc:hash-table-value 'a a3 (lambda () (error "unexpected test 1")))
      ==> 3 )

    ( (tc:hash-table-value 'a b4c5a3 (lambda () (error "unexpected test 2")))
      ==> 3 )

    ( (tc:hash-table-value 'b b4c5a3 (lambda () (error "unexpected test 3")))
      ==> 4 )

    ( (tc:hash-table-value 'c b4c5a3 (lambda () (error "unexpected test 4")))
      ==> 5 )

    )
 )

(define newtab (tc:hash-table-join
		(tc:hash-table-associate* '(b d) '(99 86) emptyhash)
		b4c5a3))


(run-regression-tests-equal?
 `(
    ( (not (not (tc:hash-table-defined? 'd newtab)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'a newtab)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'b newtab)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'c newtab)))
      ==> #t )

    ( (not (not (tc:hash-table-defined? 'e newtab)))
      ==> #f )

    ( (tc:hash-table-value 'b newtab (lambda () (error "unexpected test 5")))
      ==> 99 )

    ( (tc:hash-table-value 'd newtab (lambda () (error "unexpected test 6")))
      ==> 86 )

    ( (tc:hash-table-value 'a newtab (lambda () (error "unexpected test 7")))
      ==> 3 )

    ( (tc:hash-table-value 'c newtab (lambda () (error "unexpected test 8")))
      ==> 5 )

    ( (tc:hash-table-value 'b b4c5a3 (lambda () (error "unexpected test 9")))
      ==> 4 )
    
    ( (tc:hash-table-value 'b
			   (tc:hash-table-map
			    (lambda (key val)
			      (if (equal? key 'b) 342 val))
			    b4c5a3)
			   (lambda () (error "unexpected test 10")))
      ==> 342 )

    ( (tc:hash-table-value 'c
			   (tc:hash-table-map
			    (lambda (key val)
			      (if (equal? key 'b) 342 val))
			    b4c5a3)
			   (lambda () (error "unexpected test 11")))
      ==> 5 )

    ( (tc:hash-table-value 'c
			   (tc:hash-table-filter
			    (lambda (key val)
			      (equal? key 'c))
			    b4c5a3)
			   (lambda () (error "unexpected test 12")))
      ==> 5 )

    ( (not (not (tc:hash-table-defined? 'b
			   (tc:hash-table-filter
			    (lambda (key val)
			      (equal? key 'c))
			    b4c5a3))))
      ==> #f )

    )
 )


;(tc:hash-table-for-each
; (lambda (key val)
;   (displayln "key = " key ", val = " val))
; b4c5a3)
;
;(pretty-print emptyhash)
;(pretty-print a3)
;(pretty-print b4c5a3)
;(pretty-print newtab)
