;;; $Id: tc-scheme-scanner.scm,v 1.6 2006/02/06 01:09:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

(module tc-scheme-scanner (lib "typedscm.ss" "typedscm")
  
(provide tc:char-seq-scanner
	 tc:character-string-scanner
	 *tc:allow-tokens-not-in-R5RS*
	 tc:scheme-scanner-start-interactive-use!
	 tc:scheme-scanner-end-interactive-use!
	 tc:scheme-scanner-start-non-interactive-use!
	 tc:scheme-scanner-end-non-interactive-use!
	 ;; all-from tc-scheme-token
	 tc:scheme-token tc:scheme-token? tc:scheme-token-record
	 tc:scheme-token->data tc:scheme-token->position
	 tc:scheme-token->class tc:token-class->string
	 tc:position tc:position? tc:make-token-position tc:make-tree-position
	 tc:position->start-line tc:position->end-line tc:position->filename
	 tc:position->error-string tc:position->error-lines-string
	 tc:current-scanner-position tc:current-line tc:new-line!
	 tc:current-input-filename
	 tc:start-input-filename! tc:end-input-filename!)
	
(deftype tc:char-seq-scanner (-> (tc:char-seq) tc:token-seq))
(deftype tc:character-string-scanner (-> (string) tc:token-seq))
(deftype tc:scheme-scanner-start-interactive-use! (-> () void))
(deftype tc:scheme-scanner-end-interactive-use! (-> () void))
(deftype tc:scheme-scanner-start-interactive-use! (-> () void))
(deftype tc:scheme-scanner-end-interactive-use! (-> () void))

(deftype *tc:allow-tokens-not-in-R5RS* boolean)
(deftype *tc:carriage-return-char* char)

(require (lib "tc-scheme-token.scm" "typedscm")
	 (lib "tc-char-seq-as-stream.scm" "typedscm")
	 (lib "relax-type-checking.scm" "typedscm")
	 (lib "tc-token-seq-as-stream.scm" "typedscm"))

;;; the following can be changed for R5RS compliance
(define *tc:allow-tokens-not-in-R5RS* #t)

(define *tc:carriage-return-char* (integer->char 13))



(define *tc:scheme-scanner-interactive-stack* (list #f))
(define *tc:scheme-scanner-at-least-one-emitted-stack* (list #f))
(define *tc:scheme-scanner-excess-lparens-stack* (list 0))

(define tc:scheme-scanner-start-interactive-use!
  (lambda ()
    (set! *tc:scheme-scanner-interactive-stack*
	  (cons #t *tc:scheme-scanner-interactive-stack*))
    (set! *tc:scheme-scanner-at-least-one-emitted-stack*
	  (cons #f *tc:scheme-scanner-at-least-one-emitted-stack*))
    (set! *tc:scheme-scanner-excess-lparens-stack*
	  (cons 0 *tc:scheme-scanner-excess-lparens-stack*))))

(define tc:scheme-scanner-end-interactive-use!
  (lambda ()
    (set! *tc:scheme-scanner-interactive-stack*
	  (cdr *tc:scheme-scanner-interactive-stack*))
    (set! *tc:scheme-scanner-at-least-one-emitted-stack*
	  (cdr *tc:scheme-scanner-at-least-one-emitted-stack*))
    (set! *tc:scheme-scanner-excess-lparens-stack*
	  (cdr *tc:scheme-scanner-excess-lparens-stack*))))

(define tc:scheme-scanner-start-non-interactive-use!
  (lambda ()
    (set! *tc:scheme-scanner-interactive-stack*
	  (cons #f *tc:scheme-scanner-interactive-stack*))
    (set! *tc:scheme-scanner-at-least-one-emitted-stack*
	  (cons #f *tc:scheme-scanner-at-least-one-emitted-stack*))
    (set! *tc:scheme-scanner-excess-lparens-stack*
	  (cons 0 *tc:scheme-scanner-excess-lparens-stack*))))

(define tc:scheme-scanner-end-non-interactive-use!
  (lambda ()
    (set! *tc:scheme-scanner-interactive-stack*
	  (cdr *tc:scheme-scanner-interactive-stack*))
    (set! *tc:scheme-scanner-at-least-one-emitted-stack*
	  (cdr *tc:scheme-scanner-at-least-one-emitted-stack*))
    (set! *tc:scheme-scanner-excess-lparens-stack*
	  (cdr *tc:scheme-scanner-excess-lparens-stack*))))

(define tc:scheme-scanner-emitting!
  (lambda ()
    (if (car *tc:scheme-scanner-interactive-stack*)
	(set-car! *tc:scheme-scanner-at-least-one-emitted-stack* #t))))

(define tc:scheme-scanner-count-lparen!
  (lambda ()
    (if (car *tc:scheme-scanner-interactive-stack*)
	(set-car! *tc:scheme-scanner-excess-lparens-stack*
		 (+ (car *tc:scheme-scanner-excess-lparens-stack*) 1)))))

(define tc:scheme-scanner-count-rparen!
  (lambda ()
    (if (car *tc:scheme-scanner-interactive-stack*)
	(set-car! *tc:scheme-scanner-excess-lparens-stack*
	      (- (car *tc:scheme-scanner-excess-lparens-stack*) 1)))))

(define tc:scheme-scanner-interaction-done?
  (lambda ()
    (and (car *tc:scheme-scanner-interactive-stack*)
	 (car *tc:scheme-scanner-at-least-one-emitted-stack*)
	 (zero? (car *tc:scheme-scanner-excess-lparens-stack*)))))

;;; infrastructure, adapted from EOPL Figure E.1 : page 489, 490

(define tc:shift
  (lambda (next-action)
    (lambda (buffer char-seq)
      (next-action (cons (tc:char-seq-head char-seq) buffer)
        (tc:char-seq-tail char-seq)))))

(define tc:drop
  (lambda (next-action)
    (lambda (buffer char-seq)
      (next-action buffer (tc:char-seq-tail char-seq)))))

(deftype tc:goto-scanner-state
  (forall (s t)
    (-> ((-> (char) (-> (s tc:char-seq) t))) (-> (s tc:char-seq) t))))
(define tc:goto-scanner-state
  (lambda (state)
    (lambda (buffer char-seq)
      ((state (tc:char-seq-head char-seq)) buffer char-seq))))

(define tc:emit
  (lambda (cooker)
    (lambda (buffer char-seq)
      (tc:scheme-scanner-emitting!)
      (tc:scheme-scanner-answer (cooker (reverse buffer)) char-seq))))

;;; scheme-scanner-answers
(define-datatype tc:scheme-scanner-answer-record tc:scheme-scanner-answer?
  (tc:scheme-scanner-answer (token tc:scheme-token?)
			    (unscanned tc:char-seq?)))

;;; cooking = making a token of the appropriate class
(define tc:cook-punctuation
  (lambda (class)
    (lambda (buffer)
      (case class
	((lparen sharp-lparen) (tc:scheme-scanner-count-lparen!))
	((rparen) (tc:scheme-scanner-count-rparen!)))
      (tc:scheme-token-record (tc:current-scanner-position) class '*))))

(define tc:cook-number
  (lambda (buffer)
    (let ((num (string->number (list->string buffer))))
      (if (true? num)
	  (tc:scheme-token-record (tc:current-scanner-position) 'number num)
	  (tc:scheme-scanner-error "illegal numeric literal:"
				   (list->string buffer))))))

(define tc:cook-boolean
  (lambda (buffer)
    (tc:scheme-token-record
     (tc:current-scanner-position)
     'boolean
     (eqv? #\t (cadr buffer)))))

(define tc:cook-string
  (lambda (buffer)
    (tc:scheme-token-record
     (tc:current-scanner-position)
     'string
     (list->string buffer))))

(define tc:cook-character
  (lambda (buffer)
    (tc:scheme-token-record
     (tc:current-scanner-position)
     'character
     (if (= (length buffer) 3)
	 (caddr buffer)
	 (let ((name (list->string (cddr buffer))))
	   (cond
	    ((string-ci=? "space" name)
	     #\space)
	    ((or (string-ci=? "newline" name)
		 (and *tc:allow-tokens-not-in-R5RS*
		      (string-ci=? "nl" name))) ; in SCM
	     #\newline)
	    (*tc:allow-tokens-not-in-R5RS*
	     ;; the following are all in SCM
	     (cond
	      ((or (string-ci=? "tab" name)
		   (string-ci=? "ht" name))
	       (integer->char 9))
	      ((string-ci=? "vt" name)
	       (integer->char 11))
	      ((string-ci=? "bel" name)
	       (integer->char 7))
	      ((or (string-ci=? "null" name)
		   (string-ci=? "nul" name))
	       (integer->char 0))
	      ((and (= (string-length name) 2)
		    (char=? (string-ref name 0) #\^)
		    (char<=? #\A (string-ref name 1)))
	       (integer->char
		(+ 1
		   (- (char->integer
		       (char-upcase (string-ref name 1)))
		      (char->integer #\A)))))
	      (else
	       (tc:scheme-scanner-error
		(string-append "improper character name #\\"
			       name)))))
	    (else
	     (tc:scheme-scanner-error
	      (string-append "improper character name #\\"
			     name)))))))))

(define tc:cook-identifier
  (lambda (buffer)
    (let ((identifier
	   (string->symbol (list->string (map char-downcase buffer)))))
      (tc:scheme-token-record (tc:current-scanner-position)
			      'identifier
			      identifier))))

(deftype tc:scan-once
  (forall (s t)
    (-> ((-> (char) (-> ((list-of s) tc:char-seq) t)) tc:char-seq) t)))
(define tc:scan-once
  (lambda (start-state char-seq)
    ((tc:goto-scanner-state start-state) '() char-seq)))

;;; main public procedures

(define tc:scan-char-seq
  (lambda (start-state char-seq)
    (let ((next-answer (tc:scan-once start-state char-seq)))
      (cases tc:scheme-scanner-answer-record next-answer
        (tc:scheme-scanner-answer (token unscanned)                 
          (tc:make-token-seq-with-thunk token
            (lambda ()
              (if (eq? (tc:scheme-token->class token) 'end-marker)
                  (tc:the-null-token-seq)
                  (tc:scan-char-seq start-state unscanned)))))))))

(define tc:char-seq-scanner
  (lambda (a-char-seq)
    (tc:scan-char-seq tc:scheme-scanner-start-state a-char-seq)))
		      
(define tc:character-string-scanner
  (lambda (a-string)
    (tc:scan-char-seq tc:scheme-scanner-start-state
      (tc:string->char-seq
       (string-append a-string (string *tc:scheme-scanner-end-char*))))))

;;; The end of the input is signalled by *tc:scheme-scanner-end-char*
;;; This is supposed to be something that can't occur naturally in the input
(define *tc:scheme-scanner-end-char* (integer->char 0))
(define tc:scheme-scanner-end-char?
  (lambda (c)
    (or (eq? c (has-type-trusted char 'end-of-stream))
	(char=? c *tc:scheme-scanner-end-char*))))

(deftype tc:char->string-for-err (-> (char) string))
(define tc:char->string-for-err
  (lambda (c)
    (if (tc:scheme-scanner-end-char? c)
	"<end of file>"
	(string c))))

;;; scanner errors
(deftype tc:scheme-scanner-error (-> (string ...) poof))
(define tc:scheme-scanner-error
    (lambda messages
      (error (tc:scheme-scanner-message-format messages))))

(deftype tc:scheme-scanner-warning (-> (string ...) void))
(define tc:scheme-scanner-warning
    (lambda messages
      (displayln (tc:scheme-scanner-message-format
		  (cons "Warning:" messages)))))

(deftype tc:scheme-scanner-message-format (-> ((list-of string)) string))
(define tc:scheme-scanner-message-format
  (letrec ((blank-before-each
	    (lambda (msgs)
	      (if (null? msgs)
		  ""
		  (string-append " " (car msgs)
				 (blank-before-each (cdr msgs)))))))
    (lambda (messages)
      (string-append
       (tc:current-input-filename)
       ": line "
       (number->string (tc:current-line))
       ":"
       (blank-before-each messages)))))


;;; ******************************************************************

;;; lexical classes from R5RS
(define tc:scheme-delimiter?
  (lambda (c)
    (or (tc:scheme-scanner-end-char? c) ; addition needed to prevent errors
	(char-whitespace? c)
	(char=? c #\( )
	(char=? c #\))
	(char=? c #\")
	(char=? c #\;)
	)))

(define tc:special-initial-char?
  (lambda (c)
    (case c
      ((#\! #\$ #\% #\& #\* #\/ #\: #\< #\= #\> #\? #\^ #\_ #\~) #t)
      ;; for SXML, allowed in Chez Scheme, SCM, MzScheme
      ((#\@) *tc:allow-tokens-not-in-R5RS*)
      (else #f))))

(define tc:special-subsequent-char?
  (lambda (c)
    (case c
      ((#\+ #\- #\. #\@) #t)
      (else #f))))

;; The following is for keeping track of newlines across DOS/Mac/Unix,
;; where on DOS one has \n\r, on Mac on has just \r, and on Unix just \n.
(deftype *tc:last-was-carriage-return* boolean)
(define *tc:last-was-carriage-return* #f)

;;; adapted from EOPL (first edition) Figure E.2 : pages 491-492

(define tc:scheme-scanner-start-state
  (lambda (c)
    (if (or (tc:scheme-scanner-end-char? c)
	    (not (char=? c #\newline)))
	(set! *tc:last-was-carriage-return* #f))
    (cond
     ((tc:scheme-scanner-end-char? c)
      (tc:emit (tc:cook-punctuation 'end-marker)))
     ((char-whitespace? c)
      (if (char=? c #\newline)
	  (if (not *tc:last-was-carriage-return*)
	      (tc:new-line!)
	      (set! *tc:last-was-carriage-return* #f)))
      (if (char=? c *tc:carriage-return-char*)
	  (begin
	    (set! *tc:last-was-carriage-return* #t)
	    (tc:new-line!)))
      (if (tc:scheme-scanner-interaction-done?)
	  (tc:drop (tc:emit (tc:cook-punctuation 'end-marker)))
	  (tc:drop (tc:goto-scanner-state tc:scheme-scanner-start-state))))
     ((or (char-alphabetic? c) (tc:special-initial-char? c))
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-identifier-state)))
     ((char=? c #\#)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-sharp-state)))
     ((char-numeric? c)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
     ((or (char=? c #\+) (char=? c #\-) )
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-sign-state)))
     ((char=? c #\;)
      (tc:drop (tc:goto-scanner-state tc:scheme-scanner-comment-state)))
     ((char=? c #\")
      (tc:drop (tc:goto-scanner-state tc:scheme-scanner-string-lit-state)))
     ((char=? c #\,)
      (tc:drop (tc:goto-scanner-state tc:scheme-scanner-comma-state)))
     ((char=? c #\.)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-dot-state)))
     ((char=? c #\()
      (tc:drop (tc:emit (tc:cook-punctuation 'lparen))))
     ((char=? c #\))
      (tc:drop (tc:emit (tc:cook-punctuation 'rparen))))
     ((char=? c #\')
      (tc:drop (tc:emit (tc:cook-punctuation 'quote-char))))
     ((char=? c #\`)
      (tc:drop (tc:emit (tc:cook-punctuation 'quasiquote-char))))
     (else
      (tc:scheme-scanner-warning
       "Invalid character to scan:" (tc:char->string-for-err c))
      (tc:drop (tc:goto-scanner-state tc:scheme-scanner-start-state))))))

(define tc:scheme-scanner-identifier-state
  (lambda (c)
    (cond
      ((or (char-alphabetic? c)
	   (tc:special-initial-char? c)
	   (char-numeric? c)
	   (tc:special-subsequent-char? c)
	   (and *tc:allow-tokens-not-in-R5RS*
	       (char=? c #\#))
	   )
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-identifier-state)))
      (else (tc:emit tc:cook-identifier)))))

(define tc:scheme-scanner-sharp-state
  (lambda (c)
    (case c
      ((#\\ )
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-char-lit-state)))
      ((#\( )
       (tc:drop (tc:emit (tc:cook-punctuation 'sharp-lparen))))
      ((#\t #\f) (tc:shift (tc:emit tc:cook-boolean)))
      ((#\b #\o #\d)
       (tc:shift
	(tc:goto-scanner-state tc:scheme-scanner-number-seen-radix-state)))
      ((#\x)
       (tc:shift
	(tc:goto-scanner-state tc:scheme-scanner-number-seen-hex-state)))
      ((#\i #\e)
       (tc:shift
	(tc:goto-scanner-state tc:scheme-scanner-number-seen-exactness-state)))
      (else (tc:scheme-scanner-error
	     "Invalid character following `#':"
	     (tc:char->string-for-err c))))))

(define tc:scheme-scanner-number-seen-radix-state
  (lambda (c)
    (case c
      ((#\#)
       (tc:shift
	(tc:goto-scanner-state tc:scheme-scanner-number-get-exactness-state)))
      (else (if (char-numeric? c)
		(tc:shift
		 (tc:goto-scanner-state tc:scheme-scanner-number-state))
		(tc:scheme-scanner-error
		 "Invalid character following radix:"
		 (tc:char->string-for-err c)))))))

(define tc:scheme-scanner-number-get-exactness-state
  (lambda (c)
    (case c
      ((#\i #\e)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
      (else (tc:scheme-scanner-error
	     "Invalid character following radix and `#':"
	     (tc:char->string-for-err c))))))

(define tc:scheme-scanner-number-seen-hex-state
  (lambda (c)
    (case c
      ((#\#)
       (tc:shift
	(tc:goto-scanner-state
	 tc:scheme-scanner-number-get-hex-exactness-state)))
      ((#\a #\b #\c #\d #\e #\f)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-hex-state)))
      (else (if (char-numeric? c)
		(tc:shift
		 (tc:goto-scanner-state tc:scheme-scanner-number-hex-state))
		(tc:scheme-scanner-error
		 "Invalid character following #x:"
		 (tc:char->string-for-err c)))))))

(define tc:scheme-scanner-number-get-hex-exactness-state
  (lambda (c)
    (case c
      ((#\i #\e)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-hex-state)))
      (else (tc:scheme-scanner-error
	     "Invalid character following `#x#':"
	     (tc:char->string-for-err c))))))

(define tc:scheme-scanner-number-seen-exactness-state
  (lambda (c)
    (case c
      ((#\#)
       (tc:shift
	(tc:goto-scanner-state tc:scheme-scanner-number-get-radix-state)))
      (else (if (char-numeric? c)
		(tc:shift
		 (tc:goto-scanner-state tc:scheme-scanner-number-state))
		(tc:scheme-scanner-error
		 "Invalid character following radix:"
		 (tc:char->string-for-err c)))))))

(define tc:scheme-scanner-number-get-radix-state
  (lambda (c)
    (case c
      ((#\b #\o #\d)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
      ((#\x)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-hex-state)))
      (else (tc:scheme-scanner-error
	     "Invalid character following exactness and `#':"
	     (tc:char->string-for-err c))))))

;;; The code in tc:cook-number does careful checking, using string->number.
;;; So we have a simplified version of the lexical grammar here.
(define tc:scheme-scanner-number-state
  (lambda (c)
    (case c
     ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9
       #\+ #\- #\i #\@             ; signs and complex numbers
       #\. #\e #\s #\f #\d #\l #\# ; decimals and floats
       #\/)                        ; exact rationals
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
     (else (tc:emit tc:cook-number)))))

(define tc:scheme-scanner-number-hex-state
  (lambda (c)
    (case c
     ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9
       #\a #\b #\c #\d #\e #\f
       #\+ #\- #\i #\@             ; signs and complex numbers
       #\/)                        ; exact rationals
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-hex-state)))
     (else (tc:emit tc:cook-number)))))

(define tc:scheme-scanner-sign-state
  (lambda (c)
    (cond
     ((tc:scheme-delimiter? c)
      (tc:goto-scanner-state tc:scheme-scanner-identifier-state))
     ((char=? c #\i)
      (tc:shift (tc:emit tc:cook-number)))
     ((char-numeric? c)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
     ((or *tc:allow-tokens-not-in-R5RS*
	  (char=? c #\>)) ; needed for arrows (->) in the type checker
      (tc:goto-scanner-state tc:scheme-scanner-identifier-state))
     (else (tc:scheme-scanner-error
	     "Invalid character following + or - :"
	     (tc:char->string-for-err c))))))

(define tc:scheme-scanner-comment-state
  (lambda (c)
    (cond
      ((tc:scheme-scanner-end-char? c)
       (tc:goto-scanner-state tc:scheme-scanner-start-state))
      ((char=? c #\newline)
       (tc:new-line!)
       (tc:drop (tc:goto-scanner-state tc:scheme-scanner-start-state)))
      ((char=? c *tc:carriage-return-char*) ; for the Mac
       (set! *tc:last-was-carriage-return* #t)
       (tc:new-line!)
       (tc:drop (tc:goto-scanner-state tc:scheme-scanner-start-state)))
      (else (tc:drop
	     (tc:goto-scanner-state tc:scheme-scanner-comment-state))))))

(define tc:scheme-scanner-string-lit-state
  (lambda (c)
    (if (or (tc:scheme-scanner-end-char? c)
	    (not (char=? c #\newline)))
	(set! *tc:last-was-carriage-return* #f))
    (cond
      ((tc:scheme-scanner-end-char? c)
       (tc:scheme-scanner-error "unterminated string literal!"))
      ((char=? c #\")
       (tc:drop (tc:emit tc:cook-string)))
      ((char=? c #\\)
       (tc:drop (tc:goto-scanner-state tc:scheme-scanner-string-escape-state)))
      ((char=? c #\newline)
       (if (not *tc:last-was-carriage-return*)
	   (tc:new-line!)
	   (set! *tc:last-was-carriage-return* #f))
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-string-lit-state)))
      ((char=? c *tc:carriage-return-char*) ; for the Mac
       (set! *tc:last-was-carriage-return* #t)
       (tc:new-line!)
       (tc:shift (tc:goto-scanner-state tc:scheme-scanner-string-lit-state)))
      (else (tc:shift
	     (tc:goto-scanner-state tc:scheme-scanner-string-lit-state))))))


(define tc:scheme-scanner-string-escape-state
  (lambda (c)
    (cond
     ((or (char=? c #\") (char=? c #\\)
	  (and *tc:allow-tokens-not-in-R5RS*
	       (or (char=? c #\n) (char=? c #\t))))
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-string-lit-state)))
     (else (tc:scheme-scanner-error
	    "Invalid character following \\ in string literal:"
	    (tc:char->string-for-err c))))))

(define tc:scheme-scanner-char-lit-state
  (lambda (c)
    (cond
     ((tc:scheme-scanner-end-char? c)
      (tc:scheme-scanner-error "unfinished character literal"))
     ((char-alphabetic? c)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-char-lit-name-state)))
     ((char=? c #\^)
      (tc:shift
       (tc:goto-scanner-state tc:scheme-scanner-char-lit-control-state)))
     ((char=? c #\newline)
      (tc:new-line!)
      (tc:shift (tc:emit tc:cook-character)))
     ((char=? c *tc:carriage-return-char*) ; for the Mac
      (set! *tc:last-was-carriage-return* #t)
      (tc:new-line!)
      (tc:shift (tc:emit tc:cook-character)))
     (else (tc:shift (tc:emit tc:cook-character))))))

;;; checking for legal names is done in tc:cook-character
(define tc:scheme-scanner-char-lit-name-state
  (lambda (c)
    (cond
     ((char-alphabetic? c)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-char-lit-name-state)))
     (else (tc:emit tc:cook-character)))))

;;; SCM extension,
;;; the check for if this is allowed is done in tc:cook-character
(define tc:scheme-scanner-char-lit-control-state
  (lambda (c)
    (cond
     ((tc:scheme-delimiter? c)  ; it was #\^
      (tc:emit tc:cook-character))
     (else (tc:shift (tc:emit tc:cook-character))))))

(define tc:scheme-scanner-comma-state
  (lambda (c)
    (case c
      ((#\@) (tc:drop (tc:emit (tc:cook-punctuation 'unquote-splicing-char))))
      (else (tc:emit (tc:cook-punctuation 'unquote-char))))))

(define tc:scheme-scanner-dot-state
  (lambda (c)
    (cond
     ((char-numeric? c)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-number-state)))
     ((and *tc:allow-tokens-not-in-R5RS*
	   (or (char-alphabetic? c)
	       (tc:special-initial-char? c)
	       (tc:special-subsequent-char? c)))
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-identifier-state)))
     ((eqv? c #\.)
      (tc:shift (tc:goto-scanner-state tc:scheme-scanner-two-dot-state)))
     (else (tc:emit (tc:cook-punctuation 'dot))))))

;;; we only enter this state if *tc:allow-tokens-not-in-R5RS* is false
(define tc:scheme-scanner-two-dot-state
  (lambda (c)
    (case c
     ((#\.)
      (tc:shift (tc:emit tc:cook-identifier)))
     (else (tc:scheme-scanner-error
	    "Invalid character following `.':"
	    (tc:char->string-for-err c))))))
)
