;;; $Id: tc-type-check-and-eval-loop.scm,v 1.34 2006/02/03 06:23:17 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;;---------------------------------------------------------------------

;;; We assume that you already have loaded a file like
;;;    drscheme/tc-ignore-types-at-runtime.scm
;;; if you are using DrScheme, or, if you are using Chez Scheme, then both
;;;    chez/translate-module-syntax.scm  and
;;;    chez/tc-ignore-types-at-runtime-chez.scm .
;;; These defines macros that ignore deftypes and other parts of the
;;; type notation during the evaluation process.
;;; See the shell scripts in the bin directory for details.
;;; These files can also be used if you aren't
;;; using the type checker, but want to run files that have deftypes
;;; in them.

(module tc-type-check-and-eval-loop (lib "typedscm.ss" "typedscm")

(provide  type-check-and-eval type-check-and-eval-loop type-check-exit 
	  type-check-reset-env! type-check-add-to-users-type-env!
	  type-check-help
	  type-check-eval-even-if-error! type-check-dont-eval-if-error!
	  type-tracing-set! type-tracing-off! type-tracing-level
	  tc:type-check-and-load-file tc:type-check-and-require-file 
          tc:display-type-env)	 

;;; Load all the parsing stuff
(require (lib "tc-scheme-annotate.scm" "typedscm")
	 (lib "tc-scheme-parser.scm" "typedscm")
	 (lib "tc-scheme-unparser.scm" "typedscm")
	 (lib "tc-tracing.scm" "typedscm")
         (lib "tc-types.scm" "typedscm")
	 (lib "tc-type-translate.scm" "typedscm")
	 (lib "tc-util.scm" "typedscm")
	 (lib "tc-error-output.scm" "typedscm"))

;;; Set the following to #t to evaluate even when there is a type error.
(deftype *eval-even-if-error* boolean)
(define *eval-even-if-error* #f)

(deftype type-check-eval-even-if-error! (-> () void))
(define type-check-eval-even-if-error!
  (lambda ()
    (set! *eval-even-if-error* #t)))

(deftype type-check-dont-eval-if-error! (-> () void))
(define type-check-dont-eval-if-error!
  (lambda ()
    (set! *eval-even-if-error* #f)))

(deftype *tc:type-check-and-eval-loop-users-type-env*
  (tc:env-of symbol tc:binding))
(define *tc:type-check-and-eval-loop-users-type-env* (tc:env-empty))

(deftype tc:colon string)
(define tc:colon " : ")

(deftype tc:type-check-and-load-file
  (-> (string) void))
(define tc:type-check-and-load-file
  (lambda (path)
    ;; EFFECT: type check and then load the file named by path
    (type-check-and-eval
     (tc:character-string-form-parser
      (string-append "(load \""
		     (tc:string-replace path #\\ #\/)
		     "\")")))))

(deftype tc:type-check-and-require-file
  (-> (string) void))
(define tc:type-check-and-require-file
  (lambda (path)
    ;; EFFECT: type check and then require the file named by path
    (type-check-and-eval
     (tc:character-string-form-parser
      (string-append "(require (file \""
		     (tc:string-replace path #\\ #\/)
		     "\"))")))))

(deftype tc:type-check-and-eval-loop-helper
  (-> ((tc:env-of symbol tc:binding)) poof))
(define tc:type-check-and-eval-loop-helper 
  (lambda ()
    ;; EFFECT: prompt with prompt-str, read forms, type check, eval, write, ...
    (display "typed> ")
    (force-output)
    (tc:reset-parsed-file-cache!)
    (let* ((char-seq (tc:stream->char-seq (tc:make-input-stream)))
	   (form (if (tc:char-seq-null? char-seq)
		     (exit)
		     (tc:char-seq-form-parser char-seq))))
      (type-check-and-eval form)
      (tc:type-check-and-eval-loop-helper))))

(deftype type-check-and-eval
  (-> (tc:scheme-form) datum))
(define type-check-and-eval
  (lambda (form)
      (begin
	(if (> (type-tracing-level) 3)
	    (displayln "FORM: " form))
	(if (> (type-tracing-level) 2)
	    (displayln "UNPARSED: " (tc:unparse-scheme-form form)))
	(let ((result (tc:annotation->type (tc:scheme-top-level-annotate
				   *tc:type-check-and-eval-loop-users-type-env*
				   form))))
	  (cond
	   ((tc:error-type-expr? result)
	    (tc:output-error-message result)
	    (displayln "Skipping evaluation because of the type errors..."))
	   ;; PUT STUFF IN HERE FOR OTHER CASES
	   ;; one exp only below ...

	   ((tc:declaration-type-expr? result)
	    ;; Extend the user's environment
	    (let ((delta-env (tc:declaration-type-expr->env result)))
	      ;; Display all new bindings
	      (tc:display-type-env delta-env)
              ;; update the user's type environment
	      (set! *tc:type-check-and-eval-loop-users-type-env*
		    (tc:env-join
		     delta-env
		     *tc:type-check-and-eval-loop-users-type-env*))
	      ;; Then evaluate the form
	      (tc:just-eval-form #f form)))
							 	   
	   (else
	    (write (tc:just-eval-form (tc:error-type-expr? result) form))
	    (display tc:colon)
            ;; (tc:displayln-output-type  result)
	   (tc:displayln-output-type (tc:type-unparse result))))
	  (force-output)))))

(deftype tc:display-type-env
  (-> ((tc:env-of symbol tc:type-expr)) void))
(define tc:display-type-env
  (lambda (delta-env)
    (tc:env-for-each
     (lambda (key value)
       (display key)
       (display tc:colon)		 
       (tc:displayln-output-type (tc:type-unparse
                                  (tc:binding->type-expr value))))
     delta-env)))
  
(deftype tc:just-eval-form
  (-> (boolean tc:scheme-form) datum))
(define tc:just-eval-form
  (lambda (err-typ? form)
    ;; EFFECT: Evaluate form
    (if err-typ?
	(displayln "WARNING: this may have a type error..."))
    (eval (tc:unparse-scheme-form form) (interaction-environment))))

;;; ------------------------------------

;;; The following exits the type checker
(deftype type-check-exit (-> () poof))
(define type-check-exit
  (lambda ()
    ;; (set! error tc:orig-error)
    (newline)
    (displayln "Use `(type-check-and-eval-loop)' to restart the type checker.")
    ;; go back to the top-level.
    (abort)))

;;; The following resets the type checker's environment
(deftype type-check-reset-env! (-> () void))
(define type-check-reset-env!
  (lambda ()
    (set! *tc:type-check-and-eval-loop-users-type-env* (tc:env-empty))
    (displayln "All user-defined names have been purged from the type checker")
    ))

;;; The following adds to the type checker's environment
(deftype type-check-add-to-users-type-env!
  (-> ((list-of (pair-of symbol tc:output-type-expr))) void))
(define type-check-add-to-users-type-env!
  (lambda (delta-pre-env)
    (let ((delta-environment
	   (tc:env-update*
	    (map car delta-pre-env)
	    (map (lambda (p)
		   (tc:bind-type-expr 'let (tc:type-parse (cdr p))))
	       delta-pre-env)
	    (tc:env-empty))))
      (set! *tc:type-check-and-eval-loop-users-type-env*
	    (tc:env-join
	     delta-environment
	     *tc:type-check-and-eval-loop-users-type-env*)))))

;;; The following provides a list the public type checker procedures
(deftype type-check-help (-> () void))
(define type-check-help
  (lambda ()
    (displayln "The Polymorphic Type Checker for Scheme provides ")
    (displayln "the following public procedures:")
    (displayln "type-check-help           displays helpful information")
    (displayln "type-check-exit           leave the typed interpreter loop")
    (displayln "type-check-and-eval-loop  start a typed interpreter loop")
    (displayln "type-check-reset-env!     forget all names defined so far")
    (displayln "type-check-eval-even-if-error!  evaluate even if type errors")
    (displayln "type-check-dont-eval-if-error!  don't eval if type errors")
    ;; (newline)
    ;; (displayln "The following are mainly for use by Typedscm's developers:")
    ;; (displayln "type-tracing-set!   set tracing detail level to given value")
    ;; (displayln "type-tracing-off!   set tracing detail level to 0")
    ;; (displayln "type-tracing-level  returns the numeric level of tracing detail")
    ))

;;; adjusting error and starting the type-check-and-eval-loop below...

(define tc:orig-error error)

;;; Calling the new error procedure (see below)
;;; prints an error message and then calls *tc:type-check-and-eval-start*.
(deftype *tc:type-check-and-eval-start* (-> (datum) poof))
(define *tc:type-check-and-eval-start*
  (lambda (ignored)
    (displayln "exiting")
    (exit)))

;;; Change what error does, so that we start the type-check-and-eval loop
;;; after issuing an error message, instead of dropping out to Scheme.
(define tc:modified-error
  (lambda args
    (let ((cep
	   ;; if needed, change to (current-output-port)
	   (current-error-port)))
      (display "ERROR: " cep)
      (if (not (null? args))
	  (begin (display (car args) cep)
		 (for-each (lambda (x)
			     (display " " cep)
			     (write x cep))
			   (cdr args))
		 (newline cep)))
      (*tc:type-check-and-eval-start* 'ignored))))

(deftype abort (-> () poof))
(define abort
  (lambda ()
    (*tc:type-check-and-eval-start* 'ignored)))

(deftype tc:type-check-banner (-> () void))
(define tc:type-check-banner
  (lambda ()
    (displayln
     "Typed Scheme interpreter (by Brian Dorn, Gary T. Leavens, and Curtis Clifton)")
    (displayln "Use \"(type-check-help)\" to list options.")
    (newline)))

(deftype type-check-and-eval-loop  (-> () poof))
(define type-check-and-eval-loop 
  (lambda ()
    (tc:type-check-banner)
    (tc:type-check-and-eval-loop-helper)))
)
