;;; $Id: tc-type-translate.tst,v 1.10 2006/02/10 05:23:31 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

(require (lib "tc-types.scm" "typedscm")
	 (lib "tc-type-translate.scm" "typedscm"))

(run-regression-tests-equal?
 `(
    ( (tc:type-unparse (tc:make-variable-type-expr '?t1)) ==> (forall (T) T) )
    ( (tc:type-unparse (tc:make-basic-type-expr 'number)) ==> number )
    ( (tc:type-unparse (tc:make-basic-type-expr 'poof)) ==> poof )
    ( (tc:type-unparse (tc:make-basic-type-expr 'datum)) ==> datum )
    ( (tc:type-unparse
       (tc:make-function-type-expr
	(list (tc:make-basic-type-expr 'char))
	(tc:make-basic-type-expr 'string)))
      ==> (-> (char) string) )
    ( (tc:type-unparse
       (tc:make-function-type-expr
	(list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr 'number))
	(tc:make-basic-type-expr 'string)))
      ==> (-> (char number) string) )
    ( (tc:type-unparse
       (tc:make-function-type-expr
	(list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr '...))
	(tc:make-basic-type-expr 'string)))
      ==> (-> (char ...) string) )
    ( (tc:type-unparse
       (tc:make-intersection-type-expr
	(list
	 (tc:make-function-type-expr
	  (list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr '...))
	  (tc:make-basic-type-expr 'string))
	 (tc:make-function-type-expr
	  (list (tc:make-basic-type-expr 'datum) (tc:make-basic-type-expr '...))
	  (tc:make-basic-type-expr 'string)))))
      ==> (all-of (-> (char ...) string)
		  (-> (datum ...) string)) )
    ( (tc:type-unparse
       (tc:make-applied-type-expr
	 (tc:make-basic-type-expr 'list-of)
	 (list (tc:make-basic-type-expr 'number))))
      ==> (list-of number) )
    ( (tc:type-unparse
       (tc:make-applied-type-expr
	 (tc:make-basic-type-expr 'list-of)
	 (list (tc:make-variable-type-expr '?Twn))))
      ==> (forall (T) (list-of T)) )
    ( (tc:type-unparse
       (tc:make-function-type-expr
	(list
	 (tc:make-function-type-expr
	  (list (tc:make-variable-type-expr '?Swn))
	  (tc:make-variable-type-expr '?Twn))
	 (tc:make-applied-type-expr
	  (tc:make-basic-type-expr 'list-of)
	  (list (tc:make-variable-type-expr '?Swn))))
	(tc:make-applied-type-expr
	  (tc:make-basic-type-expr 'list-of)
	  (list (tc:make-variable-type-expr '?Twn)))))
      ==> (forall (S T) (-> ((-> (S) T) (list-of S)) (list-of T))) )
    ( (tc:type-unparse
       (tc:make-variant-record-type-expr
	(list
	 (tc:make-variant-type-decl-type-expr
	  'parallel-composition
	  (list
	   (tc:make-field-type-binding-type-expr
	    'exprs
	    (tc:make-applied-type-expr
	     (tc:make-basic-type-expr 'list-of)
	     (list (tc:make-basic-type-expr 'pi-calc-expr))))))
	 (tc:make-variant-type-decl-type-expr
	  'restriction
	  (list
	   (tc:make-field-type-binding-type-expr
	    'name
	    (tc:make-basic-type-expr 'symbol))
	   (tc:make-field-type-binding-type-expr
	    'body
	    (tc:make-basic-type-expr 'pi-calc-expr))))
	 (tc:make-variant-type-decl-type-expr
	  'null
	  '()))))
      =
      '(variant-record
	(parallel-composition (exprs (list-of pi-calc-expr)))
	(restriction (name symbol) (body pi-calc-expr))
	(null)) )
    ( (tc:type-unparse
       (tc:make-variant-record-type-expr
	(list
	 (tc:make-variant-type-decl-type-expr
	  'parallel-composition
	  (list
	   (tc:make-field-type-binding-type-expr
	    'exprs
	    (tc:make-applied-type-expr
	     (tc:make-basic-type-expr 'list-of)
	     (list (tc:new-variable-type-expr))))))
	 (tc:make-variant-type-decl-type-expr
	  'restriction
	  (list
	   (tc:make-field-type-binding-type-expr
	    'name
	    (tc:make-basic-type-expr 'symbol))
	   (tc:make-field-type-binding-type-expr
	    'body
	    (tc:make-basic-type-expr 'pi-calc-expr))))
	 (tc:make-variant-type-decl-type-expr
	  'null
	  '()))))
      =
      '(forall (T)
	 (variant-record
	  (parallel-composition (exprs (list-of T)))
	  (restriction (name symbol) (body pi-calc-expr))
	  (null))) )

    ( (tc:variable-type-expr? (tc:type-parse '(forall (T) T))) = #t )
    ( (tc:type-parse 'number) = (tc:make-basic-type-expr 'number) )
    ( (tc:type-parse 'poof) = (tc:make-basic-type-expr 'poof) )
    ( (tc:type-parse 'datum) = (tc:make-basic-type-expr 'datum) )
    ( (tc:type-parse '(-> (char) string))
      = (tc:make-function-type-expr
	  (list (tc:make-basic-type-expr 'char))
	  (tc:make-basic-type-expr 'string)) )
    ( (tc:type-parse '(-> (char number) string))
      =
      (tc:make-function-type-expr
       (list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr 'number))
       (tc:make-basic-type-expr 'string)) )
    ( (tc:type-parse '(-> (char ...) string))
      =
      (tc:make-function-type-expr
       (list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr '...))
       (tc:make-basic-type-expr 'string)) )
    ( (tc:type-parse '(all-of (-> (char ...) string)
			      (-> (datum ...) string)))
      =
      (tc:make-intersection-type-expr
       (list
	(tc:make-function-type-expr
	 (list (tc:make-basic-type-expr 'char) (tc:make-basic-type-expr '...))
	 (tc:make-basic-type-expr 'string))
	(tc:make-function-type-expr
	 (list (tc:make-basic-type-expr 'datum) (tc:make-basic-type-expr '...))
	 (tc:make-basic-type-expr 'string)))) )
    ( (tc:type-parse '(list-of number))
      =
      (tc:make-applied-type-expr
       (tc:make-basic-type-expr 'list-of)
       (list (tc:make-basic-type-expr 'number))) )
    ( (tc:type-unparse (tc:type-parse '(forall (T) (list-of T))))
      =
      '(forall (T) (list-of T)) )
    ( (tc:type-unparse
       (tc:type-parse
	'(forall (S T) (-> ((-> (S) T) (list-of S)) (list-of T)))))
      =
      '(forall (S T) (-> ((-> (S) T) (list-of S)) (list-of T))) )

    ( (tc:type-parse '(type-predicate-for number))
      =
      (tc:make-type-predicate-for-type-expr
       (tc:make-basic-type-expr 'number)) )
    ( (tc:type-parse
	 '(variant-record
	   (parallel-composition (exprs (list-of pi-calc-expr)))
	   (restriction (name symbol) (body pi-calc-expr))
	   (null)))
      =
      (tc:make-variant-record-type-expr
       (list
	(tc:make-variant-type-expr
	 'parallel-composition
	 (list
	  (tc:make-field-type-binding-type-expr
	   'exprs
	   (tc:make-applied-type-expr
	    (tc:make-basic-type-expr 'list-of)
	    (list (tc:make-basic-type-expr 'pi-calc-expr))))))
	(tc:make-variant-type-expr
	 'restriction
	 (list
	  (tc:make-field-type-binding-type-expr
	   'name
	   (tc:make-basic-type-expr 'symbol))
	  (tc:make-field-type-binding-type-expr
	   'body
	   (tc:make-basic-type-expr 'pi-calc-expr))))
	(tc:make-variant-type-expr
	 'null
	 '()))) )
    )
 )
